import ast

from deepsource.parse.constant import get_constant_identifier

from deepsource.parse.name import get_name_identifier


def get_compare_identifier(node: ast.Compare):
    # print(node, dir(node))
    # print(type(node.left), dir(node.left))
    # print(node.ops, dir(node.ops))
    left_type = type(node.left)
    match left_type:
        case ast.Name:
            left = get_name_identifier(node.left)
        case ast.Constant:
            left = get_constant_identifier(node.left)
        case _:
            raise NotImplementedError(left_type)
    ops = []
    for op in node.ops:
        op_type = type(op)
        match op_type:
            case ast.Eq:
                op = "=="
            case ast.NotEq:
                op = "!="
            case ast.Lt:
                op = "<"
            case ast.LtE:
                op = "<="
            case ast.Gt:
                op = ">"
            case ast.GtE:
                op = ">="
            case ast.Is:
                op = "is"
            case ast.IsNot:
                op = "is not"
            case ast.In:
                op = "in"
            case ast.NotIn:
                op = "not in"
            case _:
                raise NotImplementedError(op_type)
        ops.append(op)
    return left, tuple(ops)



import itertools
from collections import defaultdict

from deepsource.celery_app import app
from deepsource.util.file_operations import parse_code, get_file_content


@app.task
def scan_file(file_path):
    """
    Scan a file for suspicious content.
    """
    print(f"Scanning {file_path}.")
    imports, calls, assignments, syntax_error = parse_code(
        '\n'.join(get_file_content(file_path))
    )
    # print(imports)
    sus_imports = [
        # https://datatracker.ietf.org/doc/html/rfc4648.html#section-12
        'base64',      # Hide's strings and more.
        'cgi',         # Can be used to execute arbitrary commands.
        'logging',     # Logging Configuration uses eval()
        'pickle',      # Pickle can be used to execute arbitrary code.
        'ast',         # AST can be used to execute arbitrary code.
        'importlib',   # Importlib can be used to execute arbitrary code.
        'shelve',      # Shelve can be used to execute arbitrary code, uses pickle.
        'shlex',       # Shlex can be used to execute arbitrary code/shell commands.
        'subprocess',  # Subprocess can be used to execute arbitrary code/shell commands.
        'sys',         # sys.modules/meta_path/paths might have some risks.
        # https://docs.python.org/3/library/tempfile.html#tempfile-mktemp-deprecated
        'tempfile',
        # The XML processing modules are not secure against maliciously constructed data.
        # An attacker can abuse XML features to carry out denial of service attacks,
        # access local files, generate network connections to other machines,
        # or circumvent firewalls.
        # https://docs.python.org/3/library/xml.html#xml-vulnerabilities
        'xml',
        # https://docs.python.org/3/library/zipfile.html#zipfile-resources-limitations
        'zipfile'       # zipfile can be exploited to consume disk, memory, and CPU.
    ]
    found_imports = defaultdict(list)
    for (key, value), _import in itertools.product(imports.items(), sus_imports):
        if not _import:
            continue
        if (key and _import in key) or (value and _import in value):
            found_imports[_import].extend(tuple(_imports[0] for _imports in value))

    for key in found_imports:
        found_imports[key] = list(set(found_imports[key]))

    for _import, lines in found_imports.items():
        print(f"Suspicious import: [{_import}] @ {file_path} line(s) {lines}.")
    # print(calls)
    # print(assignments)
    if syntax_error:
        print(f"Syntax Error reading file: {file_path}.")
        return False
    return True



class UnhandledParsingException(Exception):
    pass


class UnhandledAssignException(UnhandledParsingException):
    pass


class UnhandledListCompException(UnhandledParsingException):
    pass


class UnhandledJoinedStrException(UnhandledParsingException):
    pass


class UnhandledListException(UnhandledParsingException):
    pass


class UnhandledTupleException(UnhandledParsingException):
    pass


class UnhandledGenExpException(UnhandledParsingException):
    pass


class UnhandledAssignTargetException(UnhandledParsingException):
    pass


class UnhandledCallIdentifierException(UnhandledParsingException):
    pass


class UnhandledCallArgException(UnhandledParsingException):
    pass


class UnhandledAwaitException(UnhandledParsingException):
    pass


class UnhandledBinOpException(UnhandledParsingException):
    pass


class UnhandledSubscriptIdentifierException(UnhandledParsingException):
    pass



import ast

from deepsource.parse.compare import get_compare_identifier

from deepsource.parse.ttuple import get_tuple_identifier

from deepsource.exceptions.parsing import UnhandledGenExpException

from deepsource.parse.name import get_name_identifier

from deepsource.parse.constant import get_constant_identifier


def get_genexp_identifier(node: ast.GeneratorExp) -> str:
    # print(node, dir(node))
    node_elt = node.elt
    match type(node_elt):
        case ast.Constant:
            return get_constant_identifier(node_elt)
        case ast.Name:
            return get_name_identifier(node_elt)
        case ast.Call:
            from deepsource.parse.call import get_call_identifier
            return get_call_identifier(node_elt)
        case ast.Tuple:
            return get_tuple_identifier(node_elt)
        case ast.Compare:
            return get_compare_identifier(node_elt)
        case _:
            raise UnhandledGenExpException(f"{node_elt}")



import ast

from deepsource.parse.binop import get_binop_identifier

from deepsource.parse.constant import get_constant_identifier

from deepsource.exceptions.parsing import UnhandledSubscriptIdentifierException

from deepsource.parse.name import get_name_identifier
from deepsource.parse.slice import get_slice_identifier


def get_subscript_identifier(node: ast.Subscript) -> str:
    # print(node.slice, dir(node.slice))

    node_slice = node.slice
    match type(node_slice):
        case ast.Name:
            return get_name_identifier(node_slice)
        case ast.Constant:
            return get_constant_identifier(node_slice)
        case ast.Slice:
            return get_slice_identifier(node_slice)
        case ast.Subscript:
            return get_subscript_identifier(node_slice)
        case ast.BinOp:
            return get_binop_identifier(node_slice)
        case _:
            raise UnhandledSubscriptIdentifierException(f"{node_slice}")



from time import sleep
from typing import List

from deepsource.celery_app import app
from deepsource.util.file_operations import gather_file_paths

import multiprocessing
import sys


def start_scanning(paths: List[str] = None) -> list:
    from deepsource.tasks.scan_file import scan_file
    task_ids = []
    for path in paths:
        task_id = scan_file.delay(path)
        task_ids.append(task_id)
    return task_ids


def start_worker():
    from deepsource.tasks import scan_file

    app.register_task(scan_file)

    worker = app.Worker(app=app, quiet=True)
    worker.start()


def deepsource():
    valid_extensions = ['py', ]  # '.pyc', '.pyo', '.pyw', '.pyx', '.pxd', '.pxi']

    paths = [path for path in gather_file_paths(sys.argv[1])
             if '.' in path and path.split('.')[-1] in valid_extensions]

    celery_process = multiprocessing.Process(target=start_worker, args=())
    celery_process.start()
    task_ids = start_scanning(paths)
    while task_ids:
        for task_id in task_ids:
            if task_id.ready():
                task_ids.remove(task_id)
                break
        else:
            print("Waiting for tasks to finish...")
            sleep(1)
    celery_process.terminate()

    return 0


def main() -> int:
    return deepsource()



from .assign import parse_assign
from .call import parse_call
from .imports import parse_import_from, parse_import
from .syntax import parse_ast

__all__ = [
    'parse_ast',
    'parse_assign',
    'parse_call',
    'parse_import',
    'parse_import_from'
]



import ast

from deepsource.exceptions.parsing import UnhandledBinOpException


def get_binop_identifier(node: ast.BinOp) -> str:
    match type(node.op):
        case ast.Add:
            return "+"
        case ast.Sub:
            return "-"
        case ast.Mult:
            return "*"
        case ast.Div:
            return "/"
        case ast.Mod:
            return "%"
        case ast.Pow:
            return "**"
        case ast.LShift:
            return "<<"
        case ast.RShift:
            return ">>"
        case ast.BitOr:
            return "|"
        case ast.BitXor:
            return "^"
        case ast.BitAnd:
            return "&"
        case ast.FloorDiv:
            return "//"
        case _:
            raise UnhandledBinOpException(f"{node.op}")



import ast

from deepsource.parse.llist import get_list_identifier

from deepsource.parse.aawait import get_await_identifier
from deepsource.parse.binop import get_binop_identifier
from deepsource.parse.boolop import get_boolop_identifier
from deepsource.parse.call import get_call_identifier

from deepsource.exceptions.parsing import UnhandledAssignTargetException, \
    UnhandledAssignException
from deepsource.parse.attribute import get_attribute_identifier
from deepsource.parse.constant import get_constant_identifier
from deepsource.parse.ddict import get_dict_identifier
from deepsource.parse.joined_str import get_joined_str_identifier
from deepsource.parse.listcomp import get_listcomp_identifier
from deepsource.parse.name import get_name_identifier
from deepsource.parse.subscript import get_subscript_identifier
from deepsource.parse.ttuple import get_tuple_identifier


def get_assign_value(node):
    node_value = node.value
    match type(node_value):
        case ast.Constant:
            value = get_constant_identifier(node_value)
        case ast.Name:
            value = get_name_identifier(node_value)
        case ast.List:
            value = get_list_identifier(node_value)
        case ast.ListComp:
            value = get_listcomp_identifier(node_value)
        case ast.Call:
            value = get_call_identifier(node_value)
        case ast.Await:
            value = get_await_identifier(node_value)
        case ast.Subscript:
            value = get_subscript_identifier(node_value)
        case ast.Attribute:
            value = get_attribute_identifier(node_value)
        case ast.Dict:
            value = get_dict_identifier(node_value)
        case ast.BinOp:
            value = get_binop_identifier(node_value)
        case ast.BoolOp:
            value = get_boolop_identifier(node_value)
        case ast.Tuple:
            value = get_tuple_identifier(node_value)
        case ast.IfExp:
            # TODO: has a body attr, maybe check that too. skipping for now
            value = 'ifexp'
            # print(node_value, dir(node_value))
            # raise UnhandledAssignException(f"{node.value}")
        case ast.JoinedStr:
            value = get_joined_str_identifier(node_value)
        case _:
            raise UnhandledAssignException(f"{node.value}")
    return value


def parse_assign_target(node, target, value, assignments):
    target_type = type(target)
    match target_type:
        case ast.Name:
            target_value = get_name_identifier(target)
        case ast.Attribute:
            target_value = get_attribute_identifier(target)
        case ast.Assign:
            target_value = get_assign_value(target)
        case ast.Subscript:
            target_value = get_subscript_identifier(target)
        case ast.Tuple:
            values = []
            for elt in target.elts:
                match type(elt):
                    case ast.Constant:
                        elt_value = get_constant_identifier(elt)
                    case ast.Name:
                        elt_value = get_name_identifier(elt)
                    case _:
                        raise UnhandledAssignException(f"{elt}")
                values.append(elt_value)
            target_value = tuple(values)
        case _:
            raise UnhandledAssignTargetException(f"{node}, {target_type}")
    assignments[value].append((node.lineno, target_value))


def parse_assign(node, assignments):
    # print('assign', node.targets, dir(node))
    # print(node.value, node.type_comment, dir(node.value))
    value = get_assign_value(node)

    for target in node.targets:
        parse_assign_target(node, target, value, assignments)



import ast

from deepsource.parse.name import get_name_identifier

from deepsource.exceptions.parsing import UnhandledAwaitException
from deepsource.parse.call import get_call_identifier


def get_await_identifier(node: ast.Await) -> str:
    node_value = node.value
    match type(node_value):
        case ast.Call:
            value = get_call_identifier(node_value)
        case ast.Name:
            value = get_name_identifier(node_value)
        case _:
            raise UnhandledAwaitException(f"{node_value}")
    return value



import ast
import os

from deepsource.parse.syntax import parse_ast


def gather_file_paths(directory: str) -> list[str]:
    """
    Recursively gather all file paths in a directory.
    """
    file_paths = []
    for root, dirs, files in os.walk(directory):
        file_paths.extend(os.path.join(root, file) for file in files)
    return file_paths


def get_file_content(file_path: str) -> list[str]:
    """
    Read a requirements file and return a list of all the lines.
    """
    with open(file_path, 'r') as f:
        return f.readlines()


def parse_code(content: str):
    from collections import defaultdict

    imports = defaultdict(list)
    calls = defaultdict(list)
    assignments = defaultdict(list)
    try:
        syntax_tree = ast.parse(content)
    except SyntaxError:
        return imports, calls, assignments, True

    for item in syntax_tree.body:
        parse_ast(item, imports, calls, assignments)
    for key in imports:
        imports[key] = list(set(imports[key]))
    for key in calls:
        calls[key] = list(set(calls[key]))
    for key in assignments:
        assignments[key] = list(set(assignments[key]))

    return imports, calls, assignments, False



import ast

from deepsource.parse.assign import parse_assign
from deepsource.parse.call import parse_call
from deepsource.parse.imports import parse_import_from, parse_import


def parse_ast(node, imports, calls, assignments):
    match type(node):
        case ast.FunctionDef:
            for function_node in node.body:
                parse_ast(function_node, imports, calls, assignments)
        case ast.Assign:
            parse_assign(node, assignments)
        case ast.ImportFrom:
            parse_import_from(node, imports)
        case ast.Import:
            parse_import(node, imports)
        case ast.Call:
            parse_call(node, calls)
        case _:
            # print(node, dir(node))
            for child in ast.iter_child_nodes(node):
                parse_ast(child, imports, calls, assignments)



import ast
from typing import Any

from deepsource.parse.constant import get_constant_identifier

from deepsource.parse.name import get_name_identifier


def get_dict_identifier(node: ast.Dict) -> Any:
    # print(node, dir(node))
    keys = []
    for key in node.keys:
        match type(key):
            case ast.Name:
                value = get_name_identifier(key)
            case ast.Constant:
                value = get_constant_identifier(key)
            case _:
                print(key, dir(key))
                raise NotImplementedError()
        keys.append(value)

    # for value in node.values:
    #     print(value, dir(value))
    return tuple(keys)



import ast

from deepsource.parse.ttuple import get_tuple_identifier
from deepsource.parse.joined_str import get_joined_str_identifier
from deepsource.parse.ddict import get_dict_identifier
from deepsource.parse.binop import get_binop_identifier
from deepsource.parse.genexp import get_genexp_identifier
from deepsource.parse.llist import get_list_identifier
from deepsource.parse.subscript import get_subscript_identifier

from deepsource.exceptions.parsing import UnhandledCallIdentifierException,\
    UnhandledCallArgException
from deepsource.parse.attribute import get_attribute_identifier
from deepsource.parse.constant import get_constant_identifier
from deepsource.parse.name import get_name_identifier


def get_call_identifier(node: ast.Call) -> str:
    node_value = node.func
    match type(node_value):
        case ast.Name:
            name = get_name_identifier(node_value)
        case ast.Attribute:
            name = get_attribute_identifier(node_value)
        case _:
            raise UnhandledCallIdentifierException(f"{node.func}")
    return name


def parse_args(node):
    args = []
    for arg in node.args:
        match type(arg):
            case ast.Name:
                value = get_name_identifier(arg)
            case ast.Attribute:
                value = get_attribute_identifier(arg)
            case ast.Constant:
                value = get_constant_identifier(arg)
            case ast.Call:
                value = get_call_identifier(arg)
            case ast.Lambda:
                value = 'lambda'
            case ast.Subscript:
                value = get_subscript_identifier(arg)
            case ast.BinOp:
                value = get_binop_identifier(arg)
            case ast.GeneratorExp:
                value = get_genexp_identifier(arg)
            case ast.List:
                value = get_list_identifier(arg)
            case ast.Dict:
                value = get_dict_identifier(arg)
            case ast.JoinedStr:
                value = get_joined_str_identifier(arg)
            case ast.Compare:
                value = "compare"  # get_compare_identifier(arg)
            case ast.Starred:
                value = "starred"
                # TODO: handle starred
            case ast.IfExp:
                value = "ifexp"
            case ast.UnaryOp:
                value = "unaryop"
            case ast.Tuple:
                value = get_tuple_identifier(arg)
            case _:
                raise UnhandledCallArgException(f"{arg}")
        args.append(value)
    return args


def parse_call(node, calls):
    args = (node.lineno, tuple(parse_args(node)))
    name = get_call_identifier(node)
    calls[name].append(args)
    return name, args



from celery import Celery

app = Celery('DeepSource', quiet=True)

app.conf.update(
    accept_content=['json'],
    broker_transport_options={},
    broker_url='redis://localhost:6379/2',
    result_backend='redis://localhost:6379/2',
    result_backend_transport_options={},
    result_serializer='json',
    task_serializer='json',
    quiet=True,
)



from .base import SuspiciousOperation, suspicious_operation
from .parsing import UnhandledParsingException, UnhandledAssignException,\
    UnhandledAssignTargetException

__all__ = [
    'SuspiciousOperation',
    'suspicious_operation',
    'UnhandledParsingException',
    'UnhandledAssignException',
    'UnhandledAssignTargetException'
]



import ast

from deepsource.exceptions.parsing import UnhandledBinOpException


def get_boolop_identifier(node: ast.BoolOp) -> str:
    match type(node.op):
        case ast.And:
            op = 'and'
        case ast.Or:
            op = 'or'
        case _:
            raise UnhandledBinOpException(f"{node.op}")
    return op



import ast

from deepsource.exceptions.parsing import UnhandledListCompException
from deepsource.parse.call import get_call_identifier

from deepsource.parse.name import get_name_identifier

from deepsource.parse.constant import get_constant_identifier


def get_listcomp_identifier(node: ast.ListComp) -> str:
    # values = []
    # print(dir(node_value))
    # for elt in node_value.elts:
    elt = node.elt
    match type(elt):
        case ast.Constant:
            value = get_constant_identifier(elt)
        case ast.Name:
            value = get_name_identifier(elt)
        case ast.Call:
            value = get_call_identifier(elt)
        case _:
            raise UnhandledListCompException(f"{elt}")
    return value



from os.path import exists

from setuptools import find_packages, setup

setup(
    name="deepsource",
    author="Bill Schumacher",
    author_email="william.schumacher@gmail.com",
    packages=find_packages(),
    scripts=[],
    url="https://github.com/BillSchumacher/deepsource",
    license="MIT",
    description="A python scanner, work in progress...",
    long_description=open("README.rst").read() if exists("README.rst") else "",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Topic :: Security",
    ],
    install_requires=["celery", "redis"],
    version="0.1.2",
    zip_safe=False,
)



import ast

from deepsource.parse.binop import get_binop_identifier

from deepsource.exceptions.parsing import UnhandledListException

from deepsource.parse.attribute import get_attribute_identifier


from deepsource.parse.name import get_name_identifier

from deepsource.parse.constant import get_constant_identifier


def get_list_identifier(node: ast.List) -> str:
    values = []
    for elt in node.elts:
        match type(elt):
            case ast.Constant:
                elt_value = get_constant_identifier(elt)
            case ast.Name:
                elt_value = get_name_identifier(elt)
            case ast.Call:
                from deepsource.parse.call import get_call_identifier
                elt_value = get_call_identifier(elt)
            case ast.Dict:
                from deepsource.parse.ddict import get_dict_identifier
                elt_value = get_dict_identifier(elt)
            case ast.Attribute:
                elt_value = get_attribute_identifier(elt)
            case ast.BinOp:
                elt_value = get_binop_identifier(elt)
            case _:
                raise UnhandledListException(f"{elt}")
        values.append(elt_value)
    value = tuple(values)
    return value



def parse_import(node, imports):
    for name in node.names:
        imports[name.name].append((node.lineno, '__module_import'))


def parse_import_from(node, imports):
    imports[node.module].append((node.lineno, tuple(name.name for name in node.names)))



import ast
from typing import Any

from deepsource.exceptions.parsing import UnhandledJoinedStrException
from deepsource.parse.formatted_value import get_formatted_value_identifier
from deepsource.parse.name import get_name_identifier

from deepsource.parse.constant import get_constant_identifier


def get_joined_str_identifier(node: ast.JoinedStr) -> Any:
    values = []
    for value in node.values:
        match type(value):
            case ast.Constant:
                value = get_constant_identifier(value)
            case ast.Name:
                value = get_name_identifier(value)
            case ast.FormattedValue:
                value = get_formatted_value_identifier(value)
            case _:
                print(value, dir(value))
                raise UnhandledJoinedStrException(f"{value}")
        values.append(value)
    return tuple(values)
